from loguru import logger
import pandas as pd
import numpy as np

from pcnn.util import normalize, inverse_normalize, ensure_list


class DataSet:
    def __init__(self, data: pd.DataFrame, data_kwargs: dict) -> None:

        if data_kwargs['verbose'] > 0:
            logger.info("Preparing the data...")

        # Ensure compatibility
        data_kwargs = self.ensure_columns_list(data_kwargs)
        # Need to retain the names of the columns for later use
        self.temperature_column_name = data_kwargs['temperature_column']

        # Check which inputs are needed in D
        if isinstance(data_kwargs['inputs_D'][0], list):
            inputs_D = []
            for input_D in data_kwargs['inputs_D']:
                inputs_D += input_D
            inputs_D = list(set(inputs_D))
        else:
            inputs_D = data_kwargs['inputs_D']

        # Only keep useful columns
        self.X_columns = data_kwargs['temperature_column'] + data_kwargs['power_column'] + [data_kwargs['out_column']] \
                        + inputs_D + data_kwargs['case_column'] 
        if data_kwargs['neigh_column'] is not None:
            self.X_columns += data_kwargs['neigh_column']
        
        self.Y_columns = data_kwargs['temperature_column']
        
        # Data arguments
        self.data = data[self.X_columns].copy()
        self.interval = (data.index[1] - data.index[0]).seconds / 60

        self.is_normalized = False
        self.min_ = None
        self.max_ = None

        # Get which column to normalize
        self.normalized_columns = self.get_columns_to_normalize(data_kwargs)
        
        # For compatibility with torch, need the columns indices
        data_kwargs = self.get_columns_indices(data_kwargs)

        # Sanity check after this modifications
        self.check_columns(data_kwargs)
        
        # Normalizing the inputs to D, outputs and the "case" column defining heating or cooling
        if data_kwargs['verbose'] > 0:
            logger.info("Normalizing the data...")
        self.normalize(columns=self.normalized_columns)

        # Define inputs and labels    
        self.X = self.data[self.X_columns].iloc[:-1, :].copy().values
        self.Y = self.data[self.Y_columns].iloc[1:, :].copy().values

        self.data_kwargs = data_kwargs

    def ensure_columns_list(self, data_kwargs: dict):
        data_kwargs['case_column'] = ensure_list(data_kwargs['case_column'])
        data_kwargs['neigh_column'] = ensure_list(data_kwargs['neigh_column'])
        data_kwargs['temperature_column'] = ensure_list(data_kwargs['temperature_column']) 
        data_kwargs['power_column'] = ensure_list(data_kwargs['power_column'])
        return data_kwargs
    
    def get_columns_to_normalize(self, data_kwargs: dict):
        """
        Need to normalize the inputs to D, the outputs (temperatures), 
        and the "case" column defining heating or cooling, as this is used
        in module.py and expected to be between 0.1 (cooling) or 0.9 (heating)
        """
        return [x for x in self.data.columns if x in 
                data_kwargs['temperature_column'] + data_kwargs['inputs_D'] + data_kwargs['case_column']]

    def get_columns_indices(self, data_kwargs: dict):
        """
        Torch works with tensors, so we need to know which columns is where, i.e., get their index
        """
        data_kwargs['case_column'] = [i for i,x in enumerate(self.X_columns) if x in data_kwargs['case_column']]
        data_kwargs['out_column'] = [i for i,x in enumerate(self.X_columns) if x == data_kwargs['out_column']][0]
        if data_kwargs['neigh_column'] is not None:
            data_kwargs['neigh_column'] = [i for i,x in enumerate(self.X_columns) if x in data_kwargs['neigh_column']]
        data_kwargs['temperature_column'] = [i for i,x in enumerate(self.X_columns) if x in data_kwargs['temperature_column']]
        data_kwargs['power_column'] = [i for i,x in enumerate(self.X_columns) if x in data_kwargs['power_column']]

        # Check if inputs_D is a list of lists (in the case of M-PCNNs)
        if isinstance(data_kwargs['inputs_D'][0], list):
            inputs_D = []
            for input_D in data_kwargs['inputs_D']:
                inputs_D.append([i for i,x in enumerate(self.X_columns) if x in input_D])
            data_kwargs['inputs_D'] = inputs_D
        else:
            data_kwargs['inputs_D'] = [i for i,x in enumerate(self.X_columns) if x in data_kwargs['inputs_D']]
        return data_kwargs
    
    def check_columns(self, data_kwargs: dict):
        """
        Sanity check of the columns
        """
        if data_kwargs['neigh_column'] is None:
            logger.info(f'Sanity check of the columns:\n{[(w, [self.X_columns[i] for i in x]) 
                                                          for w, x in zip(['Case', 'Room temp', 'Room power', 'Out temp'],
                                    [data_kwargs['case_column'], data_kwargs['temperature_column'], 
                                     data_kwargs['power_column'], [data_kwargs['out_column']]])]}')
        else:
            logger.info(f'Sanity check of the columns:\n{[(w, [self.X_columns[i] for i in x]) 
                                                          for w, x in zip(['Case', 'Room temp', 'Room power', 'Out temp', 'Neigh temp'],
                                    [data_kwargs['case_column'], data_kwargs['temperature_column'], 
                                     data_kwargs['power_column'], [data_kwargs['out_column']], data_kwargs['neigh_column']])]}')

        logger.info(f"Inputs used in D:\n{np.array(self.X_columns)[data_kwargs['inputs_D']]}")

    def normalize(self, data=None, columns=None):
        """
        Function to normalize the dataset, i.e. scale it by the min and max values
        The min and max of each column (sensor) is kept in memory to reverse the
        operation and to be able to apply them to other datasets

        There is an additional trick here, the data is actually scaled between
        0.1 and 0.9 instead of the classical 0-1 scaling to avoid saturation
        This is supposed to help the learning
        """
        inplace = False

        # If no data is provided, take the entire data
        if data is None:
            data = self.data

            # First, check that no normalization or standardization was already performed
            assert not self.is_normalized, "The data is already normalized!"

            # Set the normalized flag to true
            self.is_normalized = True
            inplace = True

        if columns is None:
            columns = data.columns

        # Normalize the data and recall mins and maxes
        normalized_data, min_, max_ = normalize(data[columns])
        data[columns] = normalized_data.values

        if inplace:
            self.data = data
            self.min_ = min_
            self.max_ = max_
        else:
            return data, min_, max_

    def inverse_normalize(self, data=None, min_=None, max_=None, inplace: bool = False, columns=None):
        """
        Function to reverse the normalization to get the original scales back. If no data is provided,
        The entire data is scaled back.

        Args:
            data:       Data to inverse normalize
            min_:       Min to use (takes the one of the dataset by default)
            max_:       Max to use (takes the one of the dataset by default)
            inplace:    Flag to do the operation in place or not

        Returns:
            Normalized data if wanted
        """

        if columns is None:
            columns = self.normalized_columns

        # If no data is provided, take the entire DataFrame of the DataSet
        if data is None:
            # First sanity check: the data is already normalized
            assert self.is_normalized, "The data is not normalized!"
            if inplace:
                self.is_normalized = False
                data_ = self.data
            else:
                data_ = self.data.copy()

        # Otherwise, copy the data to avoid issues
        else:
            if isinstance(data, pd.Series):
                data_ = pd.DataFrame(data=data, index=data.index, columns=[data.name])
            else:
                data_ = data.copy()

        if min_ is None:
            min_ = self.min_[data_.columns]
        if max_ is None:
            max_ = self.max_[data_.columns]

        # Inverse the normalization using the needed mins and maxes
        inverse_data = inverse_normalize(data=data_[columns], min_=min_, max_=max_)
        data[columns] = inverse_data.values

        # Return the scaled data if wanted
        if inplace:
            self.data = data
        else:
            return data
        
    def get_temperarature_min_and_range(self):
        """
        Function to get the minimum and the amplitude of room temperatures.
        This is used by the physics-inspired network to unnormalize the predictions.
        """
        return self.min_.loc[self.temperature_column_name].values,\
            (self.max_ - self.min_).loc[self.temperature_column_name].values


def prepare_data(data: pd.DataFrame, data_kwargs: dict, verbose: int = 2):
    """
    Pipeline of actions to take to prepare the data for some model.

    Args:
        data:                   DataFrame to use
        data_kwargs:            Parameters of the data
        verbose:                Verbose of the function

    Returns:
        A preprocessed dataset ready to be put into a model
    """

    # Use the custom function to load and prepare the full dataset 
    data_kwargs['verbose'] = verbose
    dataset = DataSet(data=data, data_kwargs=data_kwargs)

    assert len(dataset.X) == len(dataset.Y), "Something weird happened!"

    # Print the result and return it
    if verbose > 0:
        logger.info("Data ready!\n")

    return dataset



"""
File defining all the parameters.

If you want to modify any of these default values, don't forget to change them both in `create_parameters` and
in the individual parameters below
"""

import os
import torch.nn.functional as F


DATA_SAVE_PATH = os.path.join("saves", "data")
MODEL_SAVE_PATH = os.path.join("saves", "models")

# Create missing directories
for path in [DATA_SAVE_PATH, MODEL_SAVE_PATH]:
    if not os.path.isdir(path):
        os.mkdir(path)

def parameters(name: str = "Default_model", save_path: str = MODEL_SAVE_PATH,
                seed: int = 0, batch_size: int = 128, shuffle: bool = True, n_epochs: int = 20,
                learning_rate: float = 0.05, decrease_learning_rate:bool = True,
                heating: bool = True, cooling: bool = True, loss = F.mse_loss,
                warm_start_length: int = 12, minimum_sequence_length: int = 5, maximum_sequence_length: int = 240,
                overlapping_distance: int = 4, validation_percentage: float = 0.2, test_percentage: float = 0.1,
                learn_initial_hidden_states: bool = True, feed_input_through_nn: bool = True,
                input_nn_hidden_sizes: list = [128], lstm_hidden_size: int = 256,
                layer_norm: bool = True, lstm_num_layers: int = 1,
                output_nn_hidden_sizes: list = [128, 128], division_factor: float = 10.,
                device: str = None, verbose: int = 2, eps: float = 1e-6):
    """
    Parameters of the models

    Returns:
        name:                       Name of the model
        save_path:                  Where to save models
        seed:                       To fix the seed for reproducibility
        heating:                    Whether to use the model for the heating season
        cooling:                    Whether to use the model for the cooling season
        room_models:                Which rooms to model
        learn_initial_hidden_states:Whether to learn the initial hidden and cell states
        warm_start_length:          Length of data to warm start the model (autoregression terms required to
                                      initialize hidden and cell states
        minimum_sequence_length:    Minimum length of a prediction sequence (forward)
        maximum_sequence_length:    Maximum length of a sequence to predict
        overlapping_distance:       Distance between overlapping sequences to predict
        batch_size:                 Batch size for the training of models
        shuffle:                    Flag to shuffle the data in training or testing procedure
        n_epochs:                   Number of epochs to train the model
        learning_rate:              Learning rate of the models
        decrease_learning_rate:     Flag to adjust the learning rate while training models
        validation_percentage:      Percentage of the data to put in the validation set
        test_percentage:            Percentage of the data to put in the test set
        feed_input_through_nn:      Flag whether to preprocess the input before the LSTM
        input_nn_hidden_sizes:      Hidden sizes of the NNs processing the input
        lstm_hidden_size:           Hidden size of the LSTMs processing the input
        lstm_num_layers:            Number of layers for the LSTMs
        layer_norm:                 Flag whether to put a normalization layer after the LSTMs
        output_nn_hidden_sizes:     Hidden sizes of the NNs processing the output
        division_factor:            Factor to scale the output of the networks to ease learning
        verbose:                    Verbose of the models
        eps:                        Small value used for precision
    """

    assert cooling | heating, "At least heating or cooling needs to be true, otherwise nothing can be done"

    if feed_input_through_nn:
        assert len(input_nn_hidden_sizes) > 0, "You need to provide some hidden sizes for the input NN."
        if isinstance(input_nn_hidden_sizes, int):
            input_nn_hidden_sizes = list(input_nn_hidden_sizes)

    assert len(output_nn_hidden_sizes) > 0, "You need to provide some hidden sizes for the output NN."
    if isinstance(output_nn_hidden_sizes, int):
        output_nn_hidden_sizes = list(output_nn_hidden_sizes)

    if not isinstance(division_factor, list):
        division_factor = [division_factor]

    return dict(name=name,
                save_path=save_path,
                seed=seed,
                heating=heating,
                cooling=cooling,
                loss=loss,
                learn_initial_hidden_states=learn_initial_hidden_states,
                warm_start_length=warm_start_length,
                minimum_sequence_length=minimum_sequence_length,
                maximum_sequence_length=maximum_sequence_length,
                overlapping_distance=overlapping_distance,
                batch_size=batch_size,
                shuffle=shuffle,
                n_epochs=n_epochs,
                learning_rate=learning_rate,
                decrease_learning_rate=decrease_learning_rate,
                validation_percentage=validation_percentage,
                test_percentage=test_percentage,
                feed_input_through_nn=feed_input_through_nn,
                input_nn_hidden_sizes=input_nn_hidden_sizes,
                lstm_hidden_size=lstm_hidden_size,
                lstm_num_layers=lstm_num_layers,
                layer_norm=layer_norm,
                output_nn_hidden_sizes=output_nn_hidden_sizes,
                division_factor=division_factor,
                device=device,
                verbose=verbose,
                eps=eps)




import os
import pandas as pd
import numpy as np
from typing import Union
from loguru import logger
import torch
from contextlib import contextmanager
from timeit import default_timer

from pcnn.parameters import DATA_SAVE_PATH


def load_data(save_name: str, save_path: str = DATA_SAVE_PATH) -> pd.DataFrame:
    """
    Function to load a dataframe if it exists

    Args:
        save_name:  Name of the file to load
        save_path:  Where to load it
    """

    # Build the full path and check its existence
    full_path = os.path.join(save_path, save_name + ".csv")
    assert os.path.exists(full_path), f"The file {full_path} doesn't exist."

    # Load the data and change the index to timestamps
    data = pd.read_csv(full_path, index_col=[0])
    data.index = pd.to_datetime(data.index)

    return data


def normalize(data: pd.DataFrame):
    """
    Function to normalize the columns of a DataFrame to 0.1-0.9

    Args:
        data: The DataFrame to normalize

    Returns:
        the normalized data, the mins and the maxs
    """

    data = data.astype(float)

    # Define and save the min and max of each column
    max_ = data.max()
    min_ = data.min()

    # Little trick to handle constant data (zero std --> cannot divide)
    # Only consider non zero variance columns
    non_zero_div = np.where(max_ - min_ > 1e-10)[0]
    # If there is a zero variance column, it is actually useless since it is constant
    if len(non_zero_div) < len(data.columns):
        print(f"Warning, columns {data.columns[np.where(max_ - min_ < 1e-10)[0]].values} are constant, really useful?")

    # Scale the data between 0.1 and 0.9
    data.iloc[:, non_zero_div] = 0.8 * (data.iloc[:, non_zero_div] - min_.iloc[non_zero_div]) / (max_.iloc[non_zero_div] - min_.iloc[non_zero_div]) + 0.1

    data.iloc[:, np.where(max_ - min_ <= 1e-10)[0]] = 0.5

    return data, min_, max_


def inverse_normalize(data: Union[np.ndarray, pd.DataFrame, float], min_: Union[pd.Series, float], max_: Union[pd.Series, float]):
    """
    Function to inverse the normalization of the columns of a DataFrame to 0.1-0.9

    Args:
        data:   The DataFrame to inverse normalize
        min_:   The min values of the columns
        max_:   The max values of the columns

    Returns:
        The original data
    """

    # If the given data is a number
    if isinstance(data, float):
        # Can get back to the original scale back
        data = min_ + (data - 0.1) * (max_ - min_) / 0.8

    else:
        # Get the places where the variance is not zero and multiply back
        # (the other columns were ignored)
        non_zero_div = np.where(max_ - min_ > 1e-10)[0]

        # If the given data is an array
        if isinstance(data, np.ndarray):
            # Can get back to the original scale back
            data[:, non_zero_div] = (data[:, non_zero_div] - 0.1) * (max_[non_zero_div] - min_[non_zero_div]).values.reshape(1, -1) / 0.8
            data[:, non_zero_div] = data[:, non_zero_div] + min_[non_zero_div].values.reshape(1, -1)

        # If the given data is a DataFrame
        elif isinstance(data, pd.DataFrame):
            # Can get back to the original scale back
            data.iloc[:, non_zero_div] = (data.iloc[:, non_zero_div] - 0.1).multiply(max_.iloc[non_zero_div] - min_.iloc[non_zero_div]) / 0.8
            data.iloc[:, non_zero_div] = data.iloc[:, non_zero_div].add(min_.iloc[non_zero_div])

        else:
            raise ValueError(f"Unexpected data type {type(data)}")

    return data


def model_save_name_factory(module, model_kwargs):
    """
    Function to create helpful and somewhat unique names to easily save and load the wanted models
    This uses the starting and ending date of the data used to fit the model, as well as another
    part in "model_name" that is specific to each model, representing the model type and possibly
    some hyperparameters' choices

    Args:
        module:         Module used in the model
        model_kwargs:   Parameters of the model, see 'parameters.py'

    Returns:
        A full name to save the model as
    """

    name = os.path.join(model_kwargs["save_path"], f"{model_kwargs['name']}_{model_kwargs['seed']}_"
                                                   f"{module}")

    name += f"_{model_kwargs['warm_start_length']}_{model_kwargs['maximum_sequence_length']}"

    if (model_kwargs["heating"]) & (not model_kwargs["cooling"]):
        name += "_heating"
    elif (model_kwargs["cooling"]) & (not model_kwargs["heating"]):
        name += "_cooling"

    return name


def format_elapsed_time(tic, toc):
    """
    Small function to print the time elapsed between tic and toc in a nice manner
    """

    diff = toc - tic
    hours = int(diff // 3600)
    minutes = int((diff - 3600 * hours) // 60)
    seconds = str(int(diff - 3600 * hours - 60 * minutes))

    # Put everything in strings
    hours = str(hours)
    minutes = str(minutes)

    # Add a zero for one digit numbers for consistency
    if len(hours) == 1:
        hours = '0' + hours
    if len(minutes) == 1:
        minutes = '0' + minutes
    if len(seconds) == 1:
        seconds = '0' + seconds

    # Final nice looking print
    return f"{hours}:{minutes}:{seconds}"


def check_GPU_availability():
    if torch.cuda.is_available():
        device = torch.device("cuda:0")
        logger.info("GPU acceleration on!")
    elif torch.backends.mps.is_available():
        device = 'mps'
        logger.info("Using mps.")
    else:
        device = "cpu"
        logger.info("Using CPU.")
    return device


@contextmanager
def elapsed_timer():
    def elapser_function(start):
        yield lambda: default_timer() - start
    return elapser_function(start=default_timer())


def ensure_list(value):
    if value is not None:
        if isinstance(value, float) or isinstance(value, int) or isinstance(value, str):
            return [value]
        else:
            if isinstance(value, pd.Series):
                return value.values
            else:
                return value
    else:
        return None


def common_initialization(temperature_difference, factor, time_elapsed_hours, parameters):
    """
    Function to compute the common initialization of the physical parameters

    Args:
        temperature_difference:         List of the temperature differences per room
        factor:                         Factor causing the observed differences, also as a lit
        time_elasped:                   List of the time elapsed per room (in hours)
        max_temperature:                Maximum room temperature of the data
        min_temperature:                Minimum room temperature of the data

    Returns:
        basic initialization values to be standardized
    """

    # Need the output as an array
    temperature_difference = ensure_list(temperature_difference)
    factor = ensure_list(factor)
    # For robustness in case a Series is passed
    max_temperature = np.array(ensure_list(parameters['max_temperature'])) 
    min_temperature = np.array(ensure_list(parameters['min_temperature'])) 

    ## Heat losses approximation
    
    # T_diff_inside ~ b * T_diff_outside * time_elapsed  (or c * T_diff_neighboring_oom * time_elapsed)
    # --> b ~ T_diff_inside / T_diff_outside / time_elapsed (or c ~ T_diff_neighboring_room / T_diff_outside / time_elapsed)
    
    # or # T_diff_inside ~ a * power * time_elapsed (or T_diff_inside ~ d * power * time_elapsed)
    # --> a ~ T_diff_inside / power / time_elapsed (or d ~ T_diff_inside / power / time_elapsed)
    initial_values = np.array(temperature_difference) / np.array(factor) / np.array(time_elapsed_hours)

    # Discretization to the right interval
    initial_values = initial_values / 60 * parameters['interval_minutes'] 

    return initial_values, max_temperature, min_temperature


def initialize_heat_losses_to_outside(temperature_difference: list, temperature_difference_to_outsie: list, time_elapsed_hours: list, parameters: dict):
    """
    Function to compute the heat losses from temperature differences

    Args:
        temperature_difference:             List of the temperature differences per room
        temperature_difference_to_outsie:   List of the temperature differences per room
        time_elasped:                       List of the time elapsed per room (in hours)
        interval:                           Interval of the data (in hours)
        min_temperature:                    Minimum room temperature of the data        
        max_temperature:                    Maximum room temperature of the data

    Returns:
        initial values for 'b' 
    """

    initial_values, max_temperature, min_temperature = common_initialization(temperature_difference=temperature_difference, 
                                                                             factor=temperature_difference_to_outsie, 
                                                                             time_elapsed_hours=time_elapsed_hours, 
                                                                             parameters=parameters)
     
    # Rescale to work with normalized data since PCNN predictions are betwween 0.1 and 0.9
    initial_values = initial_values / (max_temperature - min_temperature) * 0.8

    return ensure_list(initial_values)
    

def initialize_heat_losses_to_neighbors(neighboring_rooms: list, temperature_difference: list, temperature_difference_to_neighbors: list, time_elapsed_hours: list, parameters: dict):
    """
    Function to compute the heat losses from temperature differences

    Args:
        neighboring_rooms:                      List of the pairs of neighboring rooms - if None then assumes single-zone PCNN
        temperature_difference:                 List of the degrees lost per room
        temperature_difference_to_neighbors:    List of the temperature differences per room
        time_elasped:                           List of the time elapsed per room (in hours)
        interval:                               Interval of the data (in hours)
        min_temperature:                        Minimum room temperature of the data
        max_temperature:                        Maximum room temperature of the data

    Returns:
        initial values for 'b' 
    """

    initial_values, max_temperature, min_temperature = common_initialization(temperature_difference=temperature_difference,
                                                                             factor=temperature_difference_to_neighbors,
                                                                             time_elapsed_hours=time_elapsed_hours,
                                                                             parameters=parameters)

    # Need to ensure right normalization for neighboring rooms having different temperature ranges
    # To correctly initialize 'c'
    if neighboring_rooms is not None:
        initial_values_split = [[], []]
        # Ensure we have the right number of initial values (if all the same)
        if len(initial_values) == 1:
            initial_values = list(initial_values) * len(neighboring_rooms)
        # Loop over the pairs of rooms and create two lists to have the right normalization for each room
        for i, (room1, room2) in enumerate(neighboring_rooms):
            initial_values_split[0].append(initial_values[i] / (max_temperature[room1] - min_temperature[room1]) * 0.8)
            initial_values_split[1].append(initial_values[i] / (max_temperature[room2] - min_temperature[room2]) * 0.8)
        initial_values = initial_values_split

    else:
        initial_values = initial_values / (max_temperature - min_temperature) * 0.8

    return ensure_list(initial_values)


def initialize_heat_gains_from_heating_cooling(temperature_difference: list, power: list, time_elapsed_hours: list, parameters: dict):
    """
    Function to compute the heat gains from power inputs

    Args:
        degrees_lost:             List of the degrees lost per room
        power:                    List of the power consumption per room during that time
        time_interval:            List of the time elapsed per room (in hours)
        interval:                 Interval of the data (in minutes)
        min_temperature:          Minimum room temperature of the data
        max_temperature:          Maximum room temperature of the data

    Returns:
        initial values for 'a' or 'd'
    """

    initial_values, max_temperature, min_temperature = common_initialization(temperature_difference=temperature_difference, 
                                                                              factor=power, 
                                                                              time_elapsed_hours=time_elapsed_hours, 
                                                                              parameters=parameters)

    # Rescale to work with normalized data since PCNN predictions are betwween 0.1 and 0.9
    initial_values = initial_values / (max_temperature - min_temperature) * 0.8

    # Cooling parameters must also be positive
    return np.abs(ensure_list(initial_values))


def check_initialization_physical_parameters(initial_values_physical_parameters, data_params):
    """
    Function to check the initialization of the physical parameters
    """
    assert len(initial_values_physical_parameters['a']) == len(data_params['temperature_column']), \
        f"The initial value of a is not the right size! You have {len(data_params['temperature_column'])} rooms but {len(initial_values_physical_parameters['a'])} initial values for 'a'."
    
    if data_params['outside_walls'] is not None:
        assert len(initial_values_physical_parameters['b']) == len(data_params['outside_walls']), \
            f"The initial value of b is not the right size! You have {len(data_params['outside_walls'])} external walls but {len(initial_values_physical_parameters['b'])} initial values for 'b'."
    # Single-zone PCNNs have only one 'b' parameters
    else:
        assert len(initial_values_physical_parameters['b']) == 1, \
            f"The initial value of b is not the right size! You have 1 external walls but {len(initial_values_physical_parameters['b'])} initial values for 'b'."
    
    if data_params['neighboring_rooms'] is not None:
        assert len(initial_values_physical_parameters['c']) == len(data_params['neighboring_rooms']), \
            f"The initial value of c is not the right size! You have {len(data_params['neighboring_rooms'])} pairs of rooms but {len(initial_values_physical_parameters['c'])} initial values for 'c'."
    # Single-zone PCNNs don't use 'neighboring_rooms'
    elif data_params['neigh_column'] is not None:
        assert len(initial_values_physical_parameters['c']) == len(data_params['neigh_column']), \
            f"The initial value of c is not the right size! You have {len(data_params['neigh_column'])} pairs of rooms but {len(initial_values_physical_parameters['c'])} initial values for 'c'."
    
    assert len(initial_values_physical_parameters['d']) == len(data_params['temperature_column']), \
        f"The initial value of d is not the right size! You have {len(data_params['temperature_column'])} rooms but {len(initial_values_physical_parameters['d'])} initial values for 'd'."